import asyncio
import hashlib
import os
from datetime import datetime

from playwright.async_api import async_playwright
from playwright.sync_api import sync_playwright
from jinja2 import Template
import markdown
from PIL import Image
import io

from xhm_html2image.MinIOClient import xhm_oss


def read_file(file_path):
    with open(file_path, 'r', encoding='utf-8') as file:
        return file.read()


def markdown_to_html(markdown_content):
    return markdown.markdown(markdown_content, extensions=['extra', 'codehilite'])


def render_template(template_string, context):
    template = Template(template_string)
    return template.render(context)


async def html_to_image(html_content, width=1242):  # 使用iPhone 8 Plus的宽度作为默认值
    url = None
    async with async_playwright() as p:
        browser = await p.chromium.launch(headless=True)
        page = await browser.new_page(viewport={'width': width, 'height': 100})  # 初始高度不重要

        image_file_name = hashlib.md5(html_content.encode('utf-8')).hexdigest()
        output_file = f"{image_file_name}.png"

        try:
            await page.set_content(html_content, wait_until='networkidle')
            # 使用 JavaScript 增加页面缩放比例以提高细节
            # page.evaluate('document.body.style.zoom = "1.0"')

            # 获取页面的实际高度
            height = await page.evaluate('document.documentElement.scrollHeight')
            await page.set_viewport_size({'width': width, 'height': height})

            # 截取整个页面
            screenshot = await page.screenshot(full_page=True)

            # 使用 Pillow 处理和保存图像
            image = Image.open(io.BytesIO(screenshot))
            image.save(output_file)

            date_str = datetime.now().strftime("%Y%m%d")

            object_name = xhm_oss.upload(
                object_name=xhm_oss.auto_object_name(file_path=output_file, object_dir=f"screen/{date_str}"),
                file_path=output_file)
            url = xhm_oss.get_url(object_name=object_name)

            print(f"Image saved as {output_file} url: {url}")
        finally:
            await browser.close()
            # 判断路径是否为文件
            if os.path.isfile(output_file):  # 如果是文件
                # 删除文件
                os.remove(output_file)

    return url


def mkqrcode(data: str):
    import qrcode
    import base64
    from io import BytesIO

    # 生成二维码图像
    qr = qrcode.QRCode(
        version=1,
        error_correction=qrcode.constants.ERROR_CORRECT_L,
        box_size=10,
        border=1,
    )

    # 添加数据
    qr.add_data(data)
    qr.make(fit=True)

    # 生成二维码图像
    img = qr.make_image(fill='black', back_color='white')

    # 将图像转换为Base64编码的字符串
    buffer = BytesIO()
    img.save(buffer, format="PNG")
    return base64.b64encode(buffer.getvalue()).decode('utf-8')


async def process_markdown_and_template(title: str, content: str, width=1242,
                                        author: str = "",
                                        qrcode_data: str = "",
                                        qrcode_url: str = "",
                                        template_file: str = "template_rich.html"):
    if not title or not content:
        return ""
    template_file = os.path.join(os.path.dirname(__file__), template_file)
    html_content = markdown_to_html(content)
    html_template = read_file(template_file)
    qrcode = ""
    if not qrcode_url and qrcode_data:
        qrcode = mkqrcode(qrcode_data)
    full_html = render_template(html_template,
                                {'content': html_content, "title": title, "author": author, "qrcode": qrcode,
                                 "qrcode_url": qrcode_url})
    return await html_to_image(full_html, width)


if __name__ == "__main__":
    # 使用示例
    title = "中国选手领奖台上的这个动作刷屏！网友：彰显奥林匹克的意义"
    content = """### 中国选手领奖台上的这个动作刷屏4"""
    author = """"""
    qrcode_data = "https://fs-oss.fscut.com/wechat/screen/20240806/81cd8ab53cad58104c934b5d591b14d3.png"
    qrcode_url = "https://fs-oss.fscut.com/wechat/robot/cypnest_qrcode.png"

    asyncio.run(process_markdown_and_template(title=title, content=content, author=author, qrcode_data=qrcode_data,
                                              qrcode_url=qrcode_url))



import urllib.parse


def encode(url):
    """urlencode之后，不会再二次encode，防止链接打不开"""
    # Check if the URL contains any non-ASCII characters
    if any(ord(c) >= 128 for c in url):
        # If it does, encode the URL using urllib.parse.quote
        encoded_url = urllib.parse.quote(url, safe='/:?&=', encoding='utf-8')
        return encoded_url
    else:
        # If it doesn't, return the original URL
        return url



def singleton(cls):
    instances = {}

    class SingletonWrapper(cls):
        def __new__(cls, *args, **kwargs):
            if cls not in instances:
                instances[cls] = super(SingletonWrapper, cls).__new__(cls)
                instances[cls].__init__(*args, **kwargs)
            return instances[cls]

    return SingletonWrapper



import requests
from urllib.parse import urlparse
from enum import Enum


class FileType(Enum):
    IMAGE = "Image"
    PDF = "PDF"
    VIDEO = "Video"
    AUDIO = "Audio"
    TEXT = "Text"
    ZIP_ARCHIVE = "ZIP Archive"
    WORD_DOCUMENT = "Word Document"
    EXCEL_DOCUMENT = "Excel Document"
    POWERPOINT_DOCUMENT = "PowerPoint Document"
    JSON_DOCUMENT = "JSON Document"
    APPLICATION_FILE = "Application File"
    UNKNOWN = "Unknown File Type"
    INVALID_URL = "Invalid URL"
    REQUEST_FAILED = "Request Failed"


def _is_valid_url(url):
    try:
        result = urlparse(url)
        return all([result.scheme in ["http", "https"], result.netloc])
    except Exception:
        return False


def get_url_file_type(url):
    if not _is_valid_url(url):
        return FileType.INVALID_URL
    try:
        # 发送 HEAD 请求获取响应头
        response = requests.head(url, allow_redirects=True)
        content_type = response.headers.get('Content-Type', '')

        # 检查不同的 MIME 类型
        if content_type.startswith('image/'):
            return FileType.IMAGE
        elif content_type == 'application/pdf':
            return FileType.PDF
        elif content_type.startswith('video/'):
            return FileType.VIDEO
        elif content_type.startswith('audio/'):
            return FileType.AUDIO
        elif content_type.startswith('text/'):
            return FileType.TEXT
        elif content_type in ['application/zip', 'application/x-zip-compressed']:
            return FileType.ZIP_ARCHIVE
        elif content_type in ['application/msword',
                              'application/vnd.openxmlformats-officedocument.wordprocessingml.document']:
            return FileType.WORD_DOCUMENT
        elif content_type in ['application/vnd.ms-excel',
                              'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet']:
            return FileType.EXCEL_DOCUMENT
        elif content_type == 'application/vnd.ms-powerpoint':
            return FileType.POWERPOINT_DOCUMENT
        elif content_type.startswith('application/json'):
            return FileType.JSON_DOCUMENT
        elif content_type.startswith('application/'):
            return FileType.APPLICATION_FILE
        else:
            return FileType.UNKNOWN

    except requests.RequestException:
        return FileType.REQUEST_FAILED


def is_url_of_type(url, file_type):
    if not isinstance(file_type, FileType):
        raise ValueError("file_type must be an instance of FileType Enum.")

    detected_file_type = get_url_file_type(url)
    return detected_file_type == file_type



from random import choices
from typing import Tuple, Optional

import numpy as np
from pydantic import BaseModel
from xhm_config import conf
from xhm_log import error

try:
    from redis.connection import ConnectionPool
    from xhm_redis import XHMRedis

    REDIS_ENABLED = True
except ImportError:
    REDIS_ENABLED = False


def inverse_softmax(x):
    """
     Compute inverse softmax values for each a set/list of scores x,
    for which higher score is mapped to lower value.

    """
    e_x = np.exp(np.max(x) - x)
    return e_x / e_x.sum()


class LLMConfig:
    MAX_SCORE: int = 100
    PRE_CACHE: str = "llm_config"

    @classmethod
    def get_pool(cls):
        return ConnectionPool(
            host=conf.get("xhm_config_redis.host"),
            port=conf.get("xhm_config_redis.port"),
            password=conf.get("xhm_config_redis.password"),
            db=3,
            encoding="utf-8",
            decode_responses=True,
        )

    def __init__(self):
        if REDIS_ENABLED:
            self._cache = XHMRedis(connection_pool=self.get_pool())
        else:
            self._cache = None

    def get_config_repair_key(self, sid: str, ori_keys_available: str, attempt_number=0) -> (str, str):
        # 本函数主要是优选可用的key
        api_key, score = self._get_available_key(sid=sid, attempt_number=attempt_number,
                                                 ori_keys_available=ori_keys_available)

        return api_key, str(score)

    def _get_available_key(self, sid: str, ori_keys_available: str, attempt_number: int = 0) -> Tuple[str, float]:
        """随机从key池选择一个key
        """
        # 获取所有key，注意直接加key分数不能小于0，否则不会被选择
        if REDIS_ENABLED:
            keys_available = self._cache.zrangebyscore(f"{self.PRE_CACHE}_{sid}",
                                                       min=0, max='+inf', withscores=True)
        else:
            keys_available = []

        if len(keys_available) == 0:
            # 从数据库初始化keys
            keys = ori_keys_available.split(",")
            keys_available = [(element, 1.0) for element in keys]
            if REDIS_ENABLED:
                self._cache.zadd(f"{self.PRE_CACHE}_{sid}", mapping={key: 1 for key in keys})
        if attempt_number <= 1:
            # 第一次按权重随机选
            scores = list(zip(*keys_available))[1]
            w = inverse_softmax(scores)
            return choices(keys_available, k=1, weights=w)[0]
        else:
            # retry的时候按分数升序依次选
            return keys_available[(attempt_number - 1) % len(keys_available)]

    def error_key(self, sid: str, api_key: str = "", score: str = "0", error_type=""):
        """Move API_KEY to discarded keys pool if the corresponding ERROR_TYPE implies that
                the key need to be discarded. Fill up the available keys pool if needed.

        """

        if api_key != "":
            error(f"sid: {sid} api_key:{api_key} error: {error_type}")
            # increase the score to reduce the weight when randomly choosing a key
            if float(score) < self.MAX_SCORE:
                self._cache.zincrby(f"{self.PRE_CACHE}_{sid}", 1.0, api_key)


_llm_conf = LLMConfig()


class LLMInfo(BaseModel):
    sid: str
    # OPENAI_API_KEY
    api_key: Optional[str] = ""
    # OPENAI_API_BASE
    api_base: Optional[str] = ""
    proxy: Optional[str] = ""
    api_type: Optional[str] = ""
    api_version: Optional[str] = ""
    score: Optional[str] = ""
    query: Optional[dict] = {}
    # 额外的属性,重试的次数
    attempt_number: int = 0
    keys: str = ""

    def __call__(self):
        api_key, score = _llm_conf.get_config_repair_key(sid=self.sid,
                                                         ori_keys_available=self.keys,
                                                         attempt_number=self.attempt_number)
        self.api_key = api_key
        self.score = score
        return self

    def error_key(self, error_type=""):
        if self.sid and self.api_key:
            _llm_conf.error_key(sid=self.sid, api_key=self.api_key, score=self.score, error_type=error_type)



import hashlib
import os
from datetime import timedelta

import minio
import mimetypes


class MinIOClient:
    """自建minio oss https://min.io/docs/minio/linux/index.html?ref=con"""

    def __init__(self, bucket_name: str):
        self.bucket_name = bucket_name
        self.minio_conf = self.get_minio_conf()

    def get_minio_conf(self):
        return {
            # 'endpoint': '10.1.251.87:9000',
            # 'secure': False,
            'endpoint': 'fs-oss.fscut.com',
            'secure': True,
            'access_key': 'xrAKwPcaKlmb3UPp8BXL',
            'secret_key': 'OFLjarU1m18XUOJYUlWcGyUSuZEIFOUA5GIUMYpO',
        }

    def auto_object_name(self, file_path: str, object_dir: str = ""):
        # 获取文件名和后缀
        file_dir = os.path.dirname(file_path)
        if not object_dir:
            object_dir = hashlib.md5(file_dir.encode('utf-8')).hexdigest()
        file_name = os.path.basename(file_path)  # 获得文件名，包括后缀
        # file_extension = os.path.splitext(file_name)[1]  # 获得文件后缀
        return f"{object_dir}/{file_name}" if object_dir else file_name

    def upload(self, file_path: str, object_name: str = None,
               content_type=None):

        object_name = object_name if object_name else self.auto_object_name(file_path=file_path)

        content_type = content_type if content_type else mimetypes.guess_type(file_path)[0]
        client = minio.Minio(**self.minio_conf)
        client.fput_object(bucket_name=self.bucket_name, object_name=object_name,
                           file_path=file_path,
                           content_type=content_type
                           )
        return object_name

    def get_presigned_url(self, object_name: str, expires: int = 604800):
        """获取预签名url"""
        client = minio.Minio(**self.minio_conf)
        expires = timedelta(seconds=expires)
        return client.presigned_get_object(bucket_name=self.bucket_name, object_name=object_name, expires=expires)

    def get_url(self, object_name: str):
        """需要自定义访问策略，用gpt写了一个
        {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Principal": {
                        "AWS": "*"
                    },
                    "Action": [
                        "s3:GetObject"
                    ],
                    "Resource": [
                        "arn:aws:s3:::wechat/*"
                    ]
                }
            ]
        }"""
        scheme = "https" if self.minio_conf["secure"] else "http"
        return f"{scheme}://{self.minio_conf['endpoint']}/{self.bucket_name}/{object_name}"

    def load(self):
        client = minio.Minio(**self.minio_conf)
        if not client.bucket_exists(self.bucket_name):
            return None
        data = client.get_object(self.bucket_name, 'wechat2')
        path = "receive.zip"
        with open(path, 'wb') as file_data:
            for d in data.stream(32 * 1024):
                file_data.write(d)
        return data.data


xhm_oss = MinIOClient(bucket_name="wechat")



import json
from dataclasses import dataclass, asdict
from typing import Type, TypeVar

T = TypeVar('T')


@dataclass
class ChatMessage:
    # 群组id
    room_id: str
    # 发言人id
    user_id: str
    # 发言内容
    content: str
    # 发言类型 - 语音/视频/图片/文件
    type: str
    # 内部数据，无须外部赋值，毫秒
    current_time_millis: int = 0
    # 消息唯一标识
    message_id: str = ""

    def to_json(self) -> str:
        data = asdict(self)
        data.pop("current_time_millis")
        return json.dumps(data, ensure_ascii=False)

    @classmethod
    def from_json(cls: Type[T], json_str: str, current_time_millis: int = 0) -> T:
        # 将 JSON 字符串转换为字典
        data = json.loads(json_str)
        if current_time_millis > 0:
            data["current_time_millis"] = current_time_millis
        # 返回解包后的 dataclass 对象
        return cls(**data)



import pika


class RabbitmqConsumer:
    def __init__(self, host='localhost', port=5672, user='guest', password='guest', virtual_host='/', exchange='',
                 routing_key=''):
        self.params = pika.ConnectionParameters(host=host, port=port, virtual_host=virtual_host,
                                                credentials=pika.PlainCredentials(username=user, password=password))
        self.exchange = exchange
        self.routing_key = routing_key
        self.channel = None
        self.declare_channel()

    def declare_channel(self):
        connection = pika.BlockingConnection(self.params)
        self.channel = connection.channel()
        self.channel.exchange_declare(exchange=self.exchange, exchange_type='direct')
        self.channel.queue_declare(queue=self.routing_key, durable=True)
        self.channel.queue_bind(exchange=self.exchange, routing_key=self.routing_key, queue=self.routing_key)

    def consume(self, callback):
        self.channel.basic_consume(queue=self.routing_key, on_message_callback=callback, auto_ack=True)
        self.channel.start_consuming()



from redis.client import Redis
from redis.connection import ConnectionPool


class XHMRedis(Redis):
    @classmethod
    def get_db(cls):
        return 0

    @classmethod
    def get_decode_responses(cls):
        return True

    @classmethod
    def get_pool(cls):
        return ConnectionPool(
            host="0.0.0.0",
            port=6379,
            password="",
            db=cls.get_db(),
            encoding="utf-8",
            decode_responses=cls.get_decode_responses(),
        )

    def __init__(self, connection_pool=None):
        connection_pool = connection_pool if connection_pool else self.get_pool()
        super().__init__(connection_pool=connection_pool)



import time

from xhm_chat_memory.chat_message import ChatMessage
from xhm_redis.xhm_redis import XHMRedis


class ChatMemory:

    def __init__(self, redis: XHMRedis):
        self._store = redis

    def _add(self, memory_key: str, message: ChatMessage, expire: int):
        current_time_millis = int(time.time() * 1000)
        self._store.zadd(memory_key, mapping={message.to_json(): current_time_millis})
        self._store.expire(memory_key, expire)

    @classmethod
    def _get_key(cls, message: ChatMessage, version: str):
        return f"chat:room:{version}:{message.room_id}" if message.room_id else f"chat:user:{version}:{message.user_id}"

    def add(self, messages: list[ChatMessage], expire: int = 86400, version: str = "v1"):
        for message in messages:
            memory_key = self._get_key(message=message, version=version)
            self._add(memory_key=memory_key, message=message, expire=expire)

    def get(self, message: ChatMessage, version: str = "v1", top_n: int = 40, time_period: int = 14400) -> list[
        ChatMessage]:
        """
        群聊对话记录，理论上只要近期聊的消息，4 小时以内的，超过4小时的聊天记录，基本没什么作用，还容易造成干扰
        top_n=40, 获取近40条聊天记录
        time_period=14400,单位秒，获取近4小时的聊天记录
        """
        memory_key = self._get_key(message=message, version=version)
        pipeline = self._store.pipeline()
        start_time = int(time.time() * 1000) - time_period * 1000
        end_time = int(time.time() * 1000)
        pipeline.zrevrangebyscore(name=memory_key, min=start_time, max=end_time, start=0, num=top_n, withscores=True)
        messages = pipeline.execute()[0]
        chat_memory_logs = []
        for message, current_time_millis in reversed(messages):
            chat_memory_logs.append(
                ChatMessage.from_json(json_str=message, current_time_millis=int(current_time_millis)))
        return chat_memory_logs
